"""Exposes functions and attributes related to the rendering
of templates.
"""
import os

import jinja2


environment = jinja2.Environment( # nosec
    autoescape=False,
    variable_start_string="${",
    variable_end_string="}"
)


def render(text: str, **params) -> str:
    """Renders the editable text configuration `text`."""
    t = environment.from_string(text)
    params.setdefault('env', os.environ)
    return t.render(**params)


def read(
    fn: str,
    encoding: str = 'utf-8',
    **params
) -> str:
    """Read file from `fn`."""
    with open(fn, 'rb') as f:
        return render(bytes.decode(f.read(), encoding), **params)



"""Declares primitives to parse data from the operating
system environment.
"""
import os
from typing import Callable
from typing import Mapping
from typing import Sequence

from unimatrix.exceptions import ImproperlyConfigured


def parseint(environ, key, default):
    """Parse the environment variable as an integer."""
    try:
        return int(environ.get(key) or default)
    except ValueError:
        raise ImproperlyConfigured('%s can not be cast to integer.' % key)

def parsebool(environ, key):
    """Parses the environment variable into a boolean."""
    return str.lower(environ.get(key) or '') in ('1', 'yes', 'y', 'enabled')


def parsefilepath(environ, key, default=None):
    """Parses a filepath. If the key does not exist, or the file pointed to
    by the environment variable does not exist, return ``None``.

    The `default` argument specifies a default value if the file does not
    exist at the locatio specified by the key, or the key was absent.
    """
    fp = environ.get(key)
    if fp is not None and not os.path.exists(fp):
        fp = None
    if not fp and default and os.path.exists(default):
        fp = default
    return fp


def parselist(
    environ: Mapping[str, str],
    key: str,
    sep: str = ':',
    cls: Callable[..., Sequence[str]] = tuple
) -> Sequence[str]:
    """Parses the environment variable into an iterable,
    using `sep` as a separator e.g.::

    >>> from unimatrix.lib.environ import parselist
    >>> parselist({'foo': '1:2'}, 'foo')
    ('1', '2')
    """
    values = environ.get(key) or ''
    return cls(filter(bool, str.split(values, sep)))



"""Declares :class:`PatternSet`."""
import collections.abc
import typing

from .pattern import Pattern


class PatternSet(collections.abc.Set):

    def __init__(self, patterns: set, sep: str = "."):
        self.patterns = set([Pattern(x, sep) for x in patterns])

    def __and__(self, value: set) -> set:
        return set(str(x) for x in super().__and__(value))

    def __rand__(self, value: set):
        return self & value

    def __contains__(self, value) -> bool:
        contains = None
        for pattern in self.patterns:
            if not pattern.is_match(value):
                continue
            contains = True
            break
        else:
            contains = False
        assert contains is not None # nosec
        return contains

    def __len__(self) -> int:
        return len(self.patterns)

    def __iter__(self):
        return iter(self.patterns)



# pylint: skip-file
import asyncio
import unittest
import time

from .. import security


class ConstantExecutionTestCase(unittest.TestCase):

    def execute_time(self, t):
        async def f():
            t1 = time.time()
            async with security.constant(t):
                pass
            self.assertTrue((time.time() - t1) >= t)
        asyncio.run(f())

    def test_constant_execution_2(self):
        self.execute_time(2)

    def test_constant_execution_4(self):
        self.execute_time(4)



"""Parse configuration from environment variables."""
import os

from unimatrix.lib import environ


#: Indicates that the application is running in debug mode.
DEBUG: bool = os.getenv('DEBUG') == '1'

#: The current environment in which the application is runtime. This variable
#: defaults to ``production`` to prevent unwanted side-effects when it is not
#: set.
DEPLOYMENT_ENV: str = os.getenv('DEPLOYMENT_ENV') or 'production'


#: Default number of allowed database connections to a single server, if not
#: otherwise specified.
DB_MAX_CONNECTIONS = 100
if 'DB_MAX_CONNECTIONS' in os.environ:
    DB_MAX_CONNECTIONS = int(os.getenv('DB_MAX_CONNECTIONS'))

#: Default timeout for database connections if not otherwise defined.
DB_TIMEOUT = 10
if 'DB_TIMEOUT' in os.environ:
    DB_TIMEOUT = int(os.getenv('DB_TIMEOUT'))

#: Indicates if SSL should be verified.
ENABLE_SSL = os.getenv('PYTHONHTTPSVERIFY') != '0'

#: Specifies which hosts may make requests to application, based on the
#: content of the ``Host`` header.
HTTP_ALLOWED_HOSTS: tuple = environ.parselist(
    os.environ, 'HTTP_ALLOWED_HOSTS', sep=';')

#: Indicate if Cross-Origin Resource Sharing (CORS) is enabled for the
#: application.
HTTP_CORS_ENABLED: bool = environ.parsebool(os.environ, 'HTTP_CORS_ENABLED')

#: The allowed origins as determined based on the ``Origin`` header setting
#: for Cross-Origin Resource Sharing (CORS).
HTTP_CORS_ALLOW_ORIGINS: tuple = environ.parselist(
    os.environ, 'HTTP_CORS_ALLOW_ORIGINS', sep=';')

#: The allowed request methods for cross-origin requests.
HTTP_CORS_ALLOW_METHODS: tuple = environ.parselist(
    os.environ, 'HTTP_CORS_ALLOW_METHODS', sep=':')

#: The allowed request headers setting for Cross-Origin Resource Sharing (CORS).
HTTP_CORS_ALLOW_HEADERS: tuple = environ.parselist(
    os.environ, 'HTTP_CORS_ALLOW_HEADERS', sep=':')

#: Indicates if cross-origin responses may include cookies.
HTTP_CORS_ALLOW_CREDENTIALS: bool = environ.parsebool(
    os.environ, 'HTTP_CORS_ALLOW_CREDENTIALS')

#: The list of response headers that may be included in a cross-origin
#: response, separated by a colon.
HTTP_CORS_EXPOSE_HEADERS: tuple = environ.parselist(
    os.environ, 'HTTP_CORS_EXPOSE_HEADERS', sep=':')

#: Specifies the maximum time in seconds that a browser may cache a CORS
#: response. Defaults to ``600``.
HTTP_CORS_TTL: int = environ.parseint(
    os.environ, 'HTTP_CORS_TTL', default=600)

#: Path to the URL configuration file.
HTTP_URLCONF = os.getenv('HTTP_URLCONF') or 'etc/urlconf.yml'

#: For worker-based HTTP servers, the worker timeout.
HTTP_WORKER_TIMEOUT = environ.parseint(os.environ, 'HTTP_WORKER_TIMEOUT', 180)

#: An X.509 certificate used in local development.
LOCALHOST_SSL_CRT = environ.parsefilepath(
    os.environ,'LOCALHOST_SSL_CRT', default='pki/server/noop.crt'
)

#: A private key that is used with TLS connections in local development
#: servers.
LOCALHOST_SSL_KEY = environ.parsefilepath(
    os.environ, 'LOCALHOST_SSL_KEY', default='pki/server/noop.key'
)


#: The default log level.
LOG_LEVEL = os.getenv('LOG_LEVEL') or 'WARNING'

#: The filepath to a PEM-encoded private key (either RSA or elliptic curve with
#: a supported curve) that is used by the application to sign access tokens when
#: communicating with other services. Implementations must ensure that when
#: issueing access tokens using this key, that the OpenID discovery document is
#: also served from the well-known URI (``/.well-known/openid-configuration``),
#: relative to the ``iss`` claim that it specifies on the tokens.
OAUTH2_ACTOR_KEY: str = os.getenv('OAUTH2_ACTOR_KEY')

#: Additional audiences that the server accepts.
OAUTH2_AUDIENCES: tuple = environ.parselist(
    os.environ, 'OAUTH2_AUDIENCES', sep=';'
)

#: The ``iss`` claim for tokens issued by this system.
OAUTH2_ISSUER: str = os.getenv('OAUTH2_ISSUER')

#: The list of Security Token Services (STSs) that the application trusts. The
#: services are expected to provide a ``/.well-known/openid-configuration``
#: URL, from which the JWKS holding its keys can be discovered. Multiple STSs
#: are separated by a semicolon.
OAUTH2_TRUSTED_STS: tuple = environ.parselist(
    os.environ, 'OAUTH2_TRUSTED_STS', sep=";"
)

#: The URL at which the application may exchange security tokens.
OAUTH2_TOKEN_EXCHANGE_URL = os.getenv('OAUTH2_TOKEN_EXCHANGE_URL')

#: The default secret key. This is an opaque string.
SECRET_KEY: str = os.getenv('SECRET_KEY')

# Returns the commit hash of the current application version.
VCS_COMMIT_HASH = os.getenv('VCS_COMMIT_HASH')

#: Concurrency limit in web environments.
WEB_CONCURRENCY = None
if os.getenv('WEB_CONCURRENCY'):
    WEB_CONCURRENCY = int(os.getenv('WEB_CONCURRENCY'))



# pylint: skip-file
import pytest

from .. import matches


def test_matching_no_wilcard():
    assert matches(["foo"], ["foo", "bar", "baz", "baz.taz"]) == {"foo"}
    assert matches({"foo"}, ["foo", "bar", "baz", "baz.taz"]) == {"foo"}


def test_list_matching_with_wilcard():
    patterns = ["foo", "bar.*"]
    values = ["foo", "bar", "baz", "bar.taz", "foo.bar"]
    assert matches(patterns, values) == {"foo", "bar.taz"}


def test_list_matching_with_no_matches():
    assert matches([], ["foo"]) == set()
    assert matches(["bar"], ["foo"]) == set()
    assert matches(["bar.*"], ["foo"]) == set()
    assert matches(["foo", "bar.*"], ["foo.bar"]) == set()



"""Declares common exceptions for all Unimatrix packages."""
import uuid
from typing import Any
from typing import Callable

from .lib import timezone


__all__ = [
    'CanonicalException',
    'ImproperlyConfigured',
    'ProgrammingError',
]


class ImproperlyConfigured(RuntimeError):
    """Raised when an invalid or illegal configuration was detected."""
    pass


class ProgrammingError(Exception):
    """Raised when parts of the framework are implemented incorrectly."""
    pass


class CanonicalExceptionMetaclass(type):
    http_status_code: int

    def __new__(cls, name: str, bases: tuple[type], attrs: dict[str, Any]):
        new_class = super().__new__(cls, name, bases, attrs)
        if new_class.http_status_code is not None\
        and new_class not in CanonicalException.registered:
            CanonicalException.registered.append(new_class)
        return new_class


class CanonicalException(Exception, metaclass=CanonicalExceptionMetaclass):
    """The base class for all exceptions."""
    __module__: str = 'unimatrix.exceptions'
    registered: list[CanonicalExceptionMetaclass] = []

    #: The HTTP default HTTP status code.
    http_status_code: int | None = None

    #: The default code for the exception.
    code: str = 'UNCAUGHT_EXCEPTION'

    #: The default message.
    message: str | None = None

    #: The default detail.
    detail: str | None = None

    #: The default hint
    hint: str | None = None

    #: The default log message. String formatting is applied.
    log_message: str = "Caught fatal {code} (id: {id})"

    #: The cooldown period enforced on the caller.
    backoff: dict[str, Any] | None = None

    @property
    def http_headers(self) -> dict[str, str]:
        """The HTTP headers when returning this exception to a client."""
        return self.get_http_headers()

    def __init__(
        self,
        id: str | None = None,
        code: str | None = None,
        message: str | None = None,
        detail: str | None = None,
        hint: str | None = None,
        http_status_code: int | None = None,
        **params: Any
    ):
        """Initialize a new :class:`CanonicalException`."""
        self.id = id or uuid.uuid4()
        self.code = code or self.code
        self.message = message or self.message
        self.detail = detail or self.detail
        self.hint = hint or self.hint
        self.http_status_code = http_status_code or self.http_status_code
        self.params = params
        self.backoff = None
        self.timestamp = timezone.now()

    def as_dict(self) -> dict[str, Any]:
        """Return a dictionary containing the exception properties."""
        return {
            'apiVersion': "v1",
            'kind': type(self).__name__,
            'type': "unimatrixone.io/error",
            'metadata': {
                'id': str(self.id),
                'timestamp': self.timestamp
            },
            'spec': {
                'code': self.code,
                'message': self.message,
                'detail': self.detail,
                'hint': self.hint,
                'fieldErrors': {},
            }
        }

    def get_http_headers(self) -> dict[str, str]:
        """Return a dictionary containing the HTTP headers to add to
        a client response.
        """
        return {
            'Cache-Control': "no-cache, no-store, must-revalidate",
            'Expires': "0",
            'Pragma': "no-cache",
            'X-Error-Code': self.code,
            'X-Canonical-Exception': type(self).__name__,
        }

    def log(
        self,
        func: Callable[..., Any],
        message: str | None = None
    ) -> None:
        """Use `logger` to log the exception as `level`."""
        pass

    def set_backoff(self,
        ttl: int,
        expires: int,
        attempts: int,
        timestamp: int
    ) -> None:
        """Set a backoff (cooldown) period that indicating how
        long the caller wait before retrying the request that
        caused the :class:`CanonicalException`.
        """
        self.backoff = {
            'ttl': ttl,
            'expires': expires,
            'attempts': attempts,
            'timestamp': timestamp
        }

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {self.code} (id: {self.id})>'


class UpstreamFailure(CanonicalException):
    code = 'UPSTREAM_FAILURE'
    http_status_code = 503
    message = "An upstream service caused a degradation of service quality."



# pylint: skip-file
import unittest
from unittest.mock import MagicMock
from unittest.mock import AsyncMock

from unimatrix.lib.datastructures import DTO
from unimatrix.lib.test import AsyncTestCase
from unimatrix import runtime


class RuntimeEventsTestCase(unittest.TestCase):

    def setUp(self):
        self.get_entrypoints = runtime.get_entrypoints
        runtime.IS_CONFIGURED = False

    def tearDown(self):
        runtime.get_entrypoints = self.get_entrypoints
        runtime.IS_CONFIGURED = False

    def test_setup_with_sync_function_calls_on_setup(self):
        pkg = ['foo', DTO(
            on_setup=MagicMock()
        )]
        runtime.get_entrypoints = MagicMock(return_value=[pkg])
        runtime.setup.sync()
        pkg[1].on_setup.assert_called_once()

    def test_setup_with_async_function_calls_on_setup(self):
        mock = AsyncMock()
        async def f():
            await mock()
            
        pkg = ['foo', DTO(
            on_setup=f
        )]
        runtime.get_entrypoints = MagicMock(return_value=[pkg])
        runtime.setup.sync()
        mock.assert_called_once()

    def test_setup_with_sync_function_sets_configured(self):
        runtime.setup.sync()
        self.assertTrue(runtime.IS_CONFIGURED)

    def test_teardown_with_sync_function_sets_configured(self):
        runtime.teardown.sync()
        self.assertFalse(runtime.IS_CONFIGURED)

    def test_teardown_with_sync_function_calls_on_teardown(self):
        pkg = ['foo', DTO(
            on_teardown=MagicMock()
        )]
        runtime.get_entrypoints = MagicMock(return_value=[pkg])
        runtime.teardown.sync()
        pkg[1].on_teardown.assert_called_once()

    def test_teardown_with_async_function_calls_on_teardown(self):
        mock = AsyncMock()
        async def f():
            await mock()
            
        pkg = ['foo', DTO(
            on_teardown=f
        )]
        runtime.get_entrypoints = MagicMock(return_value=[pkg])
        runtime.teardown.sync()
        mock.assert_called_once()



# pylint: skip-file
import unittest

from .. import test


class CapabilitiesDiscoveryTestCase(unittest.TestCase):

    @test.needs('never-runs')
    def test_skipped_test(self):
        self.fail("Test was not skipped.")

    @test.needs('noop')
    def test_not_skipped(self):
        pass



"""The :mod:`unimatrix.ext.etc` module provides an API
to load editable text configuration.


Installation
============
Install :mod:`unimatrix.ext.etc` using ``pip``:

.. code:: bash

  pip install --upgrade unimatrix.ext.etc


Usage
=====


Public API
==========

.. autofunction:: read
"""
from .template import read


__all__ = ['read']



"""Provides an API for date and time related functions."""
import datetime


def now():
    """Return the current date and time, represented as milliseconds since the
    UNIX epoch.
    """
    return int(datetime.datetime.now(datetime.timezone.utc).timestamp() * 1000)



# pylint: disable=line-too-long
"""Specifies the default logging configuration for Unimatrix
applications.
"""
import copy
import logging.config
import os
import sys


DEFAULT_LOG_LEVEL = 'ERROR'

DEPLOYMENT_ENV = os.getenv('DEPLOYMENT_ENV') or 'production'

DEFAULT_MSG_FMT = '%(message)s'
if DEPLOYMENT_ENV == 'local':
    DEFAULT_MSG_FMT = '%(pathname)s:%(lineno)d - %(message)s'
if not sys.stdout.isatty():
    DEFAULT_MSG_FMT = "%(levelprefix)s %(message)"

LOG_LEVEL = os.getenv('LOG_LEVEL') or 'WARNING'

SYSLOG_HOST = os.getenv('SYSLOG_HOST') or 'localhost'

SYSLOG_PORT = int(os.getenv('SYSLOG_PORT') or 5140)


LOGGING: dict = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "unimatrix.lib.logging.DefaultFormatter",
            "fmt": DEFAULT_MSG_FMT,
            "use_colors": None,
        },
        "access": {
            "()": "unimatrix.lib.logging.AccessFormatter",
            "fmt": '%(levelprefix)s %(client_addr)s - "%(request_line)s" %(status_code)s',  # noqa: E501
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
        "error": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "access": {
            "formatter": "access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "unimatrix": {"handlers": ["default"], "level": LOG_LEVEL},
        "unimatrix.error": {"handlers": ["error"], "level": "INFO"},
        "unimatrix.access": {"handlers": ["access"], "level": "INFO", "propagate": False},
    },
}


def configure(loggers=None):
    """Configure loggers."""
    defaults = copy.deepcopy(LOGGING)
    defaults['loggers'].update(loggers or {})
    logging.config.dictConfig(defaults)



# pylint: skip-file
import unittest

import yaml

from ..template import read
from ..template import render


class RenderTestCase(unittest.TestCase):

    def test_render_with_variable_from_env(self):
        value = 'foo'
        template = '${env.FOO}'
        result = render(template, env={'FOO': value})
        self.assertEqual(result, value)

    def test_render_with_variable(self):
        value = 'foo'
        template = '${foo}'
        result = render(template, foo=value)
        self.assertEqual(result, value)

    def test_read_from_filepath(self):
        etc = read(
            'etc/test_read_from_filepath.yml',
            env={'BAR': '1'}
        )
        dto = yaml.safe_load(etc)
        self.assertIn('foo', dto)
        self.assertEqual(dto['foo'], 1)



"""Exposes the settings declared in the environment variable
``UNIMATRIX_SETTINGS_MODULE``."""
import contextlib
import functools
import importlib
import inspect
import itertools
import types
import os
import sys

import unimatrix.const
import unimatrix.environ
from unimatrix.environ import DEPLOYMENT_ENV
from unimatrix.exceptions import ImproperlyConfigured


class Settings:
    """Proxy object to load settings lazily."""

    @staticmethod
    def import_settings(qualname):
        """Import `qualname` and construct a module based on the value of
        :attr:`unimatrix.environ.DEPLOYMENT_ENV`.
        """
        settings = types.ModuleType('settings')
        Settings.clone_attrs('unimatrix.environ', settings)
        if qualname is not None:
            Settings.clone_attrs(qualname, settings, True)
            Settings.clone_attrs(f'{qualname}.{DEPLOYMENT_ENV}', settings)
        return settings

    @staticmethod
    def clone_attrs(
        qualname: str,
        target: types.ModuleType,
        fatal: bool = False
    ):
        """Return a list of tuples containing all uppercased attributes of
        `module`, or those defined in the ``__all__`` member.
        """
        try:
            module = importlib.import_module(qualname)
            for k, v in inspect.getmembers(module):
                if (hasattr(module, '__all__') and k not in module.__all__)\
                or not str.isupper(k):
                    continue
                setattr(target, k, v)
        except ImportError:
            if fatal:
                raise ImproperlyConfigured(
                    "Can not import settings module %s" % qualname)


def __getattr__(attname: str):
    sys.modules[__name__].settings = Settings.import_settings(
        qualname=os.getenv('PYTHON_SETTINGS_MODULE')\
            or os.getenv('UNIMATRIX_SETTINGS_MODULE')
    )
    return sys.module[__name__].settings
